from fastapi import FastAPI, UploadFile
from fastapi.responses import RedirectResponse
from typing import List
import pprint, logging
from pathlib import Path
from inspect import Signature, Parameter
from Settings import auto_config as cfg
from datetime import date

from OntologyToAPI.core.Utility import *
from OntologyToAPI.core.Ontology import Ontology

UPLOAD_DIR = Path(cfg.UPLOAD_DIR)
UPLOAD_DIR.mkdir(parents=True, exist_ok=True)

def create_metadata_handler(connector, query: str, name: str):
    async def handler():
        try:
            result = await connector.exec_query(query)
            return result
        except Exception as e:
            logging.exception(f"Error running query for {name}")
            return {"error": str(e)}
    return handler

def create_business_model_handler(required_metadata: list, requiresParameters:dict, external_function):
    async def handler(**kwargs):
        aggregated_res = {"FromParameters": {**kwargs}, "FromMetadata": {}, "UPLOAD_DIR": UPLOAD_DIR}
        for md in required_metadata:
            pkg, name = md.name.split(":")
            try:
                aggregated_res["FromMetadata"][name] = await md.hasSource.comm_technology.exec_query(md.hasSource.query)
            except Exception as e:
                logging.exception(f"Error running query for {name}")
                return {"error": str(e)}
        result = await external_function(aggregated_res)
        return result
    params = []
    for pl, pt in requiresParameters.items(): params.append(Parameter(pl, Parameter.POSITIONAL_OR_KEYWORD, default=None, annotation=pt))
    handler.__signature__ = Signature(parameters=params)
    return handler

class APIGenerator:
    def __init__(self, showLogs: bool = False):
        self.ontology = Ontology()
        self.routes = []
        self.app = FastAPI()
        logging.basicConfig(
            level=logging.INFO if showLogs else logging.ERROR,
            format="%(asctime)s %(levelname)s %(message)s",
            datefmt="%d-%m-%Y %H:%M:%S"
        )
        logging.info(f'The {cfg.UPLOAD_DIR} directory is configured for uploaded files.')
        @self.app.get("/", include_in_schema=False)
        async def root():
            return RedirectResponse(url="/docs")

    def load_ontologies(self, paths: List[str]):
        for path in paths:
            if not Path(path).exists():
                raise FileNotFoundError(f"The path does not exist: {path}")
            self.ontology.parse_ontology(path=path)

    def serialize_ontologies(self):
        self.ontology.serialize_metadata()
        self.ontology.serialize_business_models()

    def generate_api_routes(self) -> FastAPI:
        logging.info(f'Generating API routes for the available metadata...')
        for pkg_name, pkg_data in self.ontology.data.items():
            for md in pkg_data:
                if not md.hasSource or not md.hasSource.query:
                    continue  # Skip metadata without source query
                route_path = f"/{pkg_name}/{md.name.split(':')[-1]}"
                query = md.hasSource.query
                db_connector = md.hasSource.comm_technology
                handler = create_metadata_handler(db_connector, query, md.name)
                self.app.add_api_route(
                    path=route_path,
                    endpoint=handler,
                    methods=["GET"],
                    name=md.hasSource.desc,
                    tags=[f"Metadata ({pkg_name})"],
                )
                logging.info(f'Added {md.name} API route...')
        logging.info(f'Generating API routes for the business models...')
        for bm_name, bm_dt in self.ontology.bms.items():
            name = bm_name.split(":")
            ec_func = import_function_from_file(filepath=bm_dt.externalCode.pythonFile, function_name=bm_dt.externalCode.function)
            handler = create_business_model_handler(bm_dt.requiresMetadata, bm_dt.requiresParameters, ec_func)
            self.app.add_api_route(
                path=f"/{name[0]}/{name[1]}/run",
                endpoint=handler,
                methods=["POST"],
                name=f"This endpoint executes the {bm_dt.name} business model, and retrieve it`s results.",
                tags=[f"Business Model ({bm_dt.name})"],
            )
            logging.info(f'Added {bm_name} running API route...')
        return self.app


import socket as soc

class SocketConnection:
    def __init__(self, args):
        try:
            self.socket = soc.socket(socket.AF_INET, socket.SOCK_STREAM)
            self.socket.connect((args["hasHost"], args["hasPort"]))
        except Exception as e:
            print("Connection failed:", e)

    def exec_query(self, message: str):
        try:
            self.socket.sendall(message.encode())
            return self.socket.recv(8192).decode()
        except Exception as e:
            return {'error': str(e)}


from OntologyToAPI.core.DTO.Metadata import Metadata
from OntologyToAPI.core.DTO.ExternalCode import ExternalCode

from typing import List, Optional
from pydantic import BaseModel

class BusinessModel(BaseModel):
    name: str
    requiresMetadata: Optional[List[Metadata]]
    requiresParameters: Optional[dict]
    externalCode: Optional[ExternalCode]


from rdflib import Graph, URIRef
from pathlib import Path
import logging
import pprint

logging.getLogger("rdflib").setLevel(logging.ERROR)

from OntologyToAPI.core.Queries import *
from OntologyToAPI.core.DTO.Source import *
from OntologyToAPI.core.DTO.BusinessModel import *

from OntologyToAPI.core.Connectors.IndentifyConnector import identifyConnector
from OntologyToAPI.core.Utility import ensure_package_installed

class Ontology:
    def __init__(self):
        self.g = Graph()
        self.data = {}
        self.bms = {}

    def parse_ontology(self, path: str):
        if path.endswith('.ttl'):
            self.g.parse(path, format='turtle')
        elif path.endswith('.owl') or path.endswith('.rdf'):
            self.g.parse(path, format='xml')
        else:
            raise ValueError(f'File extension must be .ttl or .owl or .rdf ({path} could not be parsed)')
        logging.info(f'Ontology \"{path}\" parsed successfully.')

    def __repr__(self):
        return (f"data(\n{pprint.pformat(self.data, indent=2)}\n)\n"
                f"business_models(\n{pprint.pformat(self.bms, indent=2)}\n)")

    def _get_metadata_by_name(self, name: str):
        for namespace, content in self.data.items():
            for metadata in content:
                if metadata.name == name:
                    return metadata
        return None

    def _verify_metadata(self, q: str):
        md = []
        for row in self.g.query(q):
            metadata_obj = self._get_metadata_by_name(self.g.qname(row[1]))
            if not metadata_obj:
                raise ValueError(
                    f'The {self.g.qname(row[1])} metadata which is required by the {self.g.qname(row[0])} could not be found in the ontology.')
            md.append(metadata_obj)
        return md

    def serialize_metadata(self):
        data = {}
        for row in self.g.query(GET_METADATA_QUERY):
            args = {self.g.qname(row[0]).split(":")[-1]: row[1] for row in
                    self.g.query(GET_COMMUNICATION_TECH_ARGS_QUERY + URIRef(row[4]) + ">)}")}
            comm_tech = identifyConnector(row[5], args)
            source = Source(desc=row[2], query=row[3], comm_technology=comm_tech)
            onto_pkg = self.g.qname(row[0]).split(":")[0]
            if onto_pkg not in data.keys(): data[onto_pkg] = []
            data[onto_pkg].append(Metadata(name=self.g.qname(row[0]), type=row[1], hasSource=source))
            logging.info(f'{self.g.qname(row[0])} METADATA serialized successfully;')
        self.data = data

    def serialize_business_models(self):
        BMs = {}
        for ec in self.g.query(GET_EXTERNAL_CODE_FOR_BM_QUERY):
            bm_name = self.g.qname(ec[0])
            required_metadata = self._verify_metadata(GET_REQUIRED_MD_FOR_BM_QUERY + URIRef(ec[0]) + ">)}")
            required_parameters = self.g.query(GET_REQUIRED_PARAMETERS_FOR_BM_QUERY + URIRef(ec[0]) + ">)}")
            for module in str(ec[3]).split(","): ensure_package_installed(module)
            func_path = Path(str(ec[2])).expanduser()
            if not func_path.is_absolute(): func_path = (Path.cwd() / func_path).resolve(strict=False)
            else: func_path = func_path.resolve(strict=False)
            if not func_path.exists() or not func_path.is_file():
                raise FileNotFoundError(f'The external code file for the {bm_name} business model could not be found at "{func_path}"')
            BMs[bm_name] = BusinessModel(name=bm_name.split(":")[-1],
                               requiresMetadata=required_metadata,
                               requiresParameters={str(l): t for _, l, t in required_parameters},
                               externalCode=ExternalCode(
                                    pythonFile=str(func_path),
                                    function=str(ec[4]),
                                    requiresLib=str(ec[3]).split(",")))
            logging.info(f'{bm_name} BUSINESS MODEL serialized successfully;')
        self.bms = BMs


import requests as req

class APIConnection:
    def __init__(self, args):
        self.baseURL = args["hasRequestURL"]

    def exec_query(self, endpoint_params: str):
        fullURL = self.baseURL + endpoint_params
        try:
            return req.get(fullURL).json()
        except Exception as e:
            return {'error': str(e)}


# import core modules
from OntologyToAPI.core.APIGenerator import APIGenerator
from OntologyToAPI.core.Ontology import Ontology
from OntologyToAPI.core.Queries import *
from OntologyToAPI.core.Utility import *


from setuptools import setup, find_packages

with open("CHANGELOG.txt", "r", encoding="utf-8") as f:
    cl = f.read()
    VERSION = cl.split("\"")[1]

classifiers = [
    'Development Status :: 3 - Alpha',
    'Intended Audience :: Developers',
    'Operating System :: Microsoft :: Windows :: Windows 11',
    'License :: OSI Approved :: MIT License',
    'Topic :: Software Development :: Libraries :: Python Modules',
    'Programming Language :: Python :: 3.11'
]

setup(
    name='ontologytoapi',
    version=VERSION,
    description='Multi purpose API Generator based on an Ontology Framework.',
    long_description=
    f"{open('README.md').read()}\n\n" +
    f"{open('CHANGELOG.txt').read()}\n\n",
    long_description_content_type="text/markdown",
    url='https://github.com/JCGCosta/OntologyToAPI',
    author='Júlio César Guimarães Costa',
    author_email='juliocesargcosta123@gmail.com',
    license='MIT License',
    classifiers=classifiers,
    keywords=['Ontology', 'API'],
    packages=find_packages(),
    install_requires=["SQLAlchemy","pydantic","requests","pymongo","rdflib","fastapi","uvicorn","aiomysql","motor","aiofiles","python-multipart"]
)


from OntologyToAPI.core.Connectors.Stateful.SocketConnection import SocketConnection
from OntologyToAPI.core.Connectors.Stateless.APIConnection import APIConnection
from OntologyToAPI.core.Connectors.Stateless.MYSQLConnection import MySQLConnection
from OntologyToAPI.core.Connectors.Stateless.MongoDBConnection import MongoDBConnection

SUPPORTED_CONNECTIONS = {
    "API": APIConnection,
    "MYSQL": MySQLConnection,
    "MONGODB": MongoDBConnection,
    "Socket": SocketConnection
}

def identifyConnector(CommunicationTechnology, args):
    try:
        connector_class = SUPPORTED_CONNECTIONS[str(CommunicationTechnology)]
        return connector_class(args)
    except KeyError as e:
        raise ValueError(f"Unsupported type or technology: {e} please check the Ontology.")